@@ -54,11 +54,58 @@ class WebHDFSRequestHandler(SimpleHTTPRequestHandler):
             self.end_headers()
             return None
 
-    def process_chunk(self, image_id, chunk, offset, extension):
+    def _slice_chunk(self, response, image_id, chunk, offset, extension, at_attempt):
+        # NB: method uses plain raise; should only be called from an except block!
+        if response.status_code == 416:
+            # TODO: return a 416 response for applicable upload issues rather than a 500 (HANSKEN-17174)
+            resource_size = response.headers['Content-Range']
+            resource_size = re.match(r'^bytes .+/(?P<size>\d+)$', resource_size, re.IGNORECASE)
+            if not resource_size:
+                log.warn('no usable resource size available in 416 response: {}', response.headers['Content-Range'])
+                raise
+            # reported size does not match offset, skip the difference
+            skip = int(resource_size.group('size')) - offset
+        elif response.status_code == 500:
+            response_text = response.text.strip()
+            # using the plain text error message to determine how much of the bytes were processed previously
+            # despite the error response
+            offsets = re.match(r'^expected offset (\d+) but was (\d+)$', response_text, re.IGNORECASE)
+            if offsets:
+                expected, sent = offsets.groups()
+                # skip the difference, send the trailing end of the buffer we tried to append
+                skip = int(expected) - int(sent)
+            else:
+                log.warn('failed to recognize offsets for partial retry from error message "{}"', response_text)
+                # raise original error with original traceback (causing a 500, see note below)
+                raise
+        else:
+            # cannot deal with this error in a satisfiable way, continue original error
+            raise
+
+        if skip < 0:
+            # not something we can solve by skipping bytes, continue original error
+            raise
+
+        # offset + skip should equal expected, return (recursive) result (not resetting the attempt
+        # counter) or bubble an error
+        log.warn('retrying partial chunk at offset {}, skipping {} bytes that seem to be present at offset {}',
+                 offset + skip, skip, offset)
+        return self.process_chunk(image_id, chunk[skip:], offset + skip, extension, at_attempt=at_attempt)
+
+    def process_chunk(self, image_id, chunk, offset, extension, at_attempt=None):
         def should_retry():
             return self.server.max_retries is None or attempt <= self.server.max_retries + 1
 
-        attempt = 1
+        def should_slice(response):
+            return (
+                # trigger 1: response 416 (range not satisfiable)
+                response.status_code == 416 or (
+                    # trigger 2: response 500 (internal error) with a specific error message body
+                    response.status_code == 500 and response.text.strip().lower().startswith('expected offset')
+                )
+            )
+
+        attempt = at_attempt or 1
         while attempt == 1 or should_retry():
             try:
                 return self.server.upload_callback(image_id=image_id, extension=extension,
@@ -74,10 +121,13 @@ class WebHDFSRequestHandler(SimpleHTTPRequestHandler):
                     time.sleep(self.server.retry_wait)
             except requests.HTTPError as e:
                 attempt += 1
+                response_text = e.response.text.strip()
+                if should_slice(e.response):
+                    return self._slice_chunk(e.response, image_id, chunk, offset, extension, at_attempt=attempt)
                 if e.response.status_code == 500 and should_retry():
                     log.warn('uploading a chunk at offset {} failed due to {}: {}: {}, retrying in {}s',
                              offset,
-                             e.response.status_code, e.response.reason, e.response.text.strip(),
+                             e.response.status_code, e.response.reason, response_text,
                              self.server.retry_wait,
                              e)
                     time.sleep(self.server.retry_wait)
@@ -86,7 +136,7 @@ class WebHDFSRequestHandler(SimpleHTTPRequestHandler):
                     log.exception('uploading a chunk at offset {} failed: {}: {}: {}',
                                   offset,
                                   # log detailed error information, including error response body
-                                  e.response.status_code, e.response.reason, e.response.text.strip(),
+                                  e.response.status_code, e.response.reason, response_text,
                                   e)
                     # raise original error with original traceback (causing a 500, see note below)
                     raise
